use rand::distributions::StandardNormal;
use rand::{thread_rng, Rng};

use std::fmt::{Display, Formatter, Result};

pub struct AnnealingResult {
    x: f64,
    y: f64,
    nb_iterations: usize,
    final_temp: f64,
}

impl Display for AnnealingResult {
    fn fmt(&self, f: &mut Formatter) -> Result {
        write!(
            f,
            "Solution found in {} iterations\n\nX={}\nY={}\nT={}",
            self.nb_iterations, self.x, self.y, self.final_temp
        )
    }
}

pub fn test_annealing() -> AnnealingResult {
    const EPSILON: f64 = 1e-10;

    let f = |x: f64| {
        10.0 * ((0.3 * x) * (1.3 * x.powi(2) + 1e-5 * x.powi(4) + 0.2 * x + 80.0).sin()).sin()
    };

    let voisin = |x: f64| {
        let picked: f64 = thread_rng().sample(StandardNormal);

        x + picked
    };

    let decrease_temp = |t: f64| 0.99 * t;

    let accept = |delta: f64, t: f64| {
        if delta >= 0.0 {
            let a = (-delta / t).exp();


            return thread_rng().gen_range(0.0, 1.0) < a;
        }

        true
    };

    let mut x: f64 = 0.0;
    let mut t: f64 = 100.0;
    let n: usize = 100;

    let mut nb_iterations: usize = 0;

    while (f(x) + 10.0).abs() >= EPSILON {
        for _ in 0..n {
            let y: f64 = voisin(x);

            let delta: f64 = f(y) - f(x);

            if accept(delta, t) {
                x = y;
            }
        }

        nb_iterations += n;
        t = decrease_temp(t);
    }

    AnnealingResult {
        x: x,
        y: f(x),
        nb_iterations: nb_iterations,
        final_temp: t,
    }
}
